/*
 *  OpenVPN-GUI -- A Windows GUI for OpenVPN.
 *
 *  Copyright (C) 2004 Mathias Sundman <mathias@nilings.se>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program (see the file COPYING included with this
 *  distribution); if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Large parts of this sourcefile is taken from openvpnserv.c from
 *  the OpenVPN source, with approval from the author, James Yonan
 *  <jim@yonan.net>.
 */


#include <windows.h>
#include "openvpn.h"
#include <stdlib.h>
#include <stdio.h>
#include <process.h>
#include "tray.h"

/* bool definitions */
#define bool int
#define true 1
#define false 0

/* These are new for 2000/XP, so they aren't in the mingw headers yet */
#ifndef BELOW_NORMAL_PRIORITY_CLASS
#define BELOW_NORMAL_PRIORITY_CLASS 0x00004000
#endif
#ifndef ABOVE_NORMAL_PRIORITY_CLASS
#define ABOVE_NORMAL_PRIORITY_CLASS 0x00008000
#endif

static HANDLE exit_event[MAX_CONFIGS];
static HANDLE hProcess[MAX_CONFIGS];

/* Registry values */
char exe_path[MAX_PATH];
char config_dir[MAX_PATH];
char ext_string[16];
char log_dir[MAX_PATH];
char priority_string[64];
char append_string[2];
char log_viewer[MAX_PATH];

struct process_info ps_info[MAX_CONFIGS];

/* Config-filename list */
char config_files[MAX_CONFIGS][MAX_PATH];
int num_configs=0; /* Number of config-files we have read. */
BOOL connect_status[MAX_CONFIGS];	/* true=connected  false=not connected */
char log_path[MAX_CONFIGS][MAX_PATH];

//struct process_info {
//  HANDLE hProcess;
//  int config;
//};

/* clear an object */
#define CLEAR(x) memset(&(x), 0, sizeof(x))

/* snprintf with guaranteed null termination */
#define mysnprintf(out, args...) \
        { \
           snprintf (out, sizeof(out), args); \
           out [sizeof (out) - 1] = '\0'; \
        }

/* write error to event log */
#define ERR(args...) \
        { \
           char x_msg[256]; \
           mysnprintf (x_msg, args); \
	   WriteLog(x_msg); \
        }

/* get a registry string */
#define QUERY_REG_STRING(name, data) \
  { \
    len = sizeof (data); \
    status = RegQueryValueEx(openvpn_key, name, NULL, &type, (byte *) data, &len); \
    if (status != ERROR_SUCCESS || type != REG_SZ) \
      { \
        SetLastError (status); \
        ERR (error_format, name); \
	RegCloseKey (openvpn_key); \
	goto finish; \
      } \
  }

#define SET_REG_STRING(name, data) \
  { \
    if(RegSetValueEx(hkResult, \
                  name, \
                  0, \
                  REG_SZ, \
                  data, \
                  strlen(data)+1)) \
    { \
      ERR("Error setting registry key \"%s\".", keyValue); \
      return(false); \
    } \
  }

void
WriteLog(char *msg)
{
FILE *fd;
  MessageBox(NULL, msg, "OpenVPN-GUI", MB_OK);
  //if( (fd=fopen("openvpn.log", "a")) == 0) return;
  //fprintf(fd, "%s\n", msg);
  //fclose(fd);
}

static bool
match (const WIN32_FIND_DATA *find, const char *ext)
{
  int i;

  if (find->dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
    return false;

  if (!strlen (ext))
    return true;

  i = strlen (find->cFileName) - strlen (ext) - 1;
  if (i < 1)
    return false;

  return find->cFileName[i] == '.' && !strcmp (find->cFileName + i + 1, ext);
}

/*
 * Modify the extension on a filename.
 */
static bool
modext (char *dest, int size, const char *src, const char *newext)
{
  int i;

  if (size > 0 && (strlen (src) + 1) <= size)
    {
      strcpy (dest, src);
      dest [size - 1] = '\0';
      i = strlen (dest);
      while (--i >= 0)
	{
	  if (dest[i] == '\\')
	    break;
	  if (dest[i] == '.')
	    {
	      dest[i] = '\0';
	      break;
	    }
	}
      if (strlen (dest) + strlen(newext) + 2 <= size)
	{
	  strcat (dest, ".");
	  strcat (dest, newext);
	  return true;
	}
      dest [0] = '\0';
    }
  return false;
}

int
BuildFileList(HANDLE hwnd)
{
  WIN32_FIND_DATA find_obj;
  HANDLE find_handle;
  BOOL more_files;
  char find_string[MAX_PATH];
  int i;
  char log_file[MAX_PATH];
  
  mysnprintf (find_string, "%s\\*", config_dir);

  find_handle = FindFirstFile (find_string, &find_obj);
  if (find_handle == INVALID_HANDLE_VALUE)
    {
      ERR ("Cannot get configuration file list using: %s", find_string);
      goto finish;
    }

  /*
   * Loop over each config file
   */
  do {
    if (num_configs >= MAX_CONFIGS) {
      ERR("OpenVPN GUI does not support more that %d configs. Please contact " \
           "the author if you have the need for more.", MAX_CONFIGS)
      break;
    }

    /* does file have the correct type and extension? */
    if (match (&find_obj, ext_string))
      {
        strncpy(config_files[num_configs], find_obj.cFileName, sizeof(config_files[num_configs]));

        /* get log file pathname */
        if (!modext (log_file, sizeof (log_file), config_files[num_configs], "log"))
          {
            ERR ("Cannot construct logfile name based on: %s", config_files[num_configs]);
            goto finish;
          }
        mysnprintf (log_path[num_configs], "%s\\%s", log_dir, log_file);


        num_configs++;
      }
      
    /* more files to process? */
    more_files = FindNextFile (find_handle, &find_obj);
  } while (more_files);
    
  FindClose (find_handle);
  
  if (num_configs == 0) {
    char msg[256];
    mysnprintf(msg, "No configuration files were found in %s. " \
                    "Please start OpenVPN-GUI again when you have created one.", \
                    config_dir);
    MessageBox(NULL, msg, "OpenVPN-GUI", MB_OK);
    return(false);
  }


  /* Clear connect_state table */
  for(i=0; i < MAX_CONFIGS; i++) 
    connect_status[i]=false;

  return(true);

finish:
  return(false);
}


int
GetRegistryKeys()
{
  /*
   * Read info from registry in key HKLM\SOFTWARE\OpenVPN
   */
  {
    HKEY openvpn_key;
    LONG status;
    DWORD len;
    DWORD type;
    char error_string[256];
    static const char error_format[]="Error querying registry key of type REG_SZ: " \
                                     "HKLM\\SOFTWARE\\OpenVPN-GUI\\%s";

    status = RegOpenKeyEx(
			  HKEY_LOCAL_MACHINE,
			  "SOFTWARE\\OpenVPN-GUI",
			  0,
			  KEY_READ,
			  &openvpn_key);

    if (status != ERROR_SUCCESS)
      {
        if(!SetRegistryKeys())
          return(false);
        
        status = RegOpenKeyEx(
			  HKEY_LOCAL_MACHINE,
			  "SOFTWARE\\OpenVPN-GUI",
			  0,
			  KEY_READ,
			  &openvpn_key);

        if (status != ERROR_SUCCESS)
          {
            ERR("Error opening registry key HKLM\\SOFTWARE\\OpenVPN-GUI.");
            return(false);
          }
      }

    /* get path to openvpn.exe */
    QUERY_REG_STRING ("exe_path", exe_path);

    /* get path to configuration directory */
    QUERY_REG_STRING ("config_dir", config_dir);

    /* get extension on configuration files */
    QUERY_REG_STRING ("config_ext", ext_string);

    /* get path to log directory */
    QUERY_REG_STRING ("log_dir", log_dir);

    /* get priority for spawned OpenVPN subprocesses */
    QUERY_REG_STRING ("priority", priority_string);

    /* should we truncate or append to logfile? */
    QUERY_REG_STRING ("log_append", append_string);

    /* get path to our log-viewer */
    QUERY_REG_STRING ("log_viewer", log_viewer);
    RegCloseKey (openvpn_key);
  }

  return(true);

finish:
  return(false);
}

int SetRegistryKeys()
{
  HKEY hkResult=NULL;
  DWORD dwDispos;
  char keyValue[100];

  if (RegCreateKeyEx(HKEY_LOCAL_MACHINE,
                   "Software\\OpenVPN-GUI",
                   0,
                   "",
                   REG_OPTION_NON_VOLATILE,
                   KEY_READ | KEY_WRITE,
                   NULL,
                   &hkResult,
                   &dwDispos) != ERROR_SUCCESS)
    {
      ERR("Error creating HKLM\\SOFTWARE\\OpenVPN-GUI key.");
      return(false);
    }  

  SET_REG_STRING("config_dir", "C:\\Program Files\\OpenVPN\\config");
  SET_REG_STRING("config_ext", "ovpn");
  SET_REG_STRING("exe_path", "C:\\Program Files\\OpenVPN\\bin\\openvpn.exe");
  SET_REG_STRING("log_dir", "C:\\Program Files\\OpenVPN\\log");
  SET_REG_STRING("log_append", "0");
  SET_REG_STRING("priority", "NORMAL_PRIORITY_CLASS");
  SET_REG_STRING("log_viewer", "C:\\windows\\notepad.exe");

  return(true);
}

void StartOpenVPN(int config)
{


  DWORD priority;
  bool append;

  HANDLE hThread; 
  DWORD IDThread; 
  char exit_event_name[50];
  
  
  /*
   * Create our exit event
   */
  exit_event[config] = NULL;
  mysnprintf(exit_event_name, "openvpn_exit_event_%d",config);
  exit_event[config] = CreateEvent (NULL, TRUE, FALSE, exit_event_name);
  if (exit_event[config] == NULL)
    {
      ERR ("CreateEvent failed on exit event: %s", exit_event_name);
      goto finish;
    }

  /* set process priority */
  priority = NORMAL_PRIORITY_CLASS;
  if (!strcmp (priority_string, "IDLE_PRIORITY_CLASS"))
    priority = IDLE_PRIORITY_CLASS;
  else if (!strcmp (priority_string, "BELOW_NORMAL_PRIORITY_CLASS"))
    priority = BELOW_NORMAL_PRIORITY_CLASS;
  else if (!strcmp (priority_string, "NORMAL_PRIORITY_CLASS"))
    priority = NORMAL_PRIORITY_CLASS;
  else if (!strcmp (priority_string, "ABOVE_NORMAL_PRIORITY_CLASS"))
    priority = ABOVE_NORMAL_PRIORITY_CLASS;
  else if (!strcmp (priority_string, "HIGH_PRIORITY_CLASS"))
    priority = HIGH_PRIORITY_CLASS;
  else
    {
      ERR ("Unknown priority name: %s", priority_string);
      goto finish;
    }
  /* set log file append/truncate flag */
  append = false;
  if (append_string[0] == '0')
    append = false;
  else if (append_string[0] == '1')
    append = true;
  else
    {
      ERR ("Log file append flag (given as '%s') must be '0' or '1'", append_string);
      goto finish;
    }

  HANDLE log_handle = NULL;
  STARTUPINFO start_info;
  PROCESS_INFORMATION proc_info;
  SECURITY_ATTRIBUTES sa;
  SECURITY_DESCRIPTOR sd;
  char command_line[256];

  CLEAR (start_info);
  CLEAR (proc_info);
  CLEAR (sa);
  CLEAR (sd);

      

  /* construct command line */
  mysnprintf (command_line, "openvpn --service %s 1 --config \"%s\"",
	      exit_event_name,
	      config_files[config]);

  /* Make security attributes struct for logfile handle so it can
     be inherited. */
  sa.nLength = sizeof (sa);
  sa.lpSecurityDescriptor = &sd;
  sa.bInheritHandle = TRUE;
  if (!InitializeSecurityDescriptor (&sd, SECURITY_DESCRIPTOR_REVISION))
    {
      ERR ("InitializeSecurityDescriptor failed");
      goto finish;
    }
  if (!SetSecurityDescriptorDacl (&sd, TRUE, NULL, FALSE))
    {
      ERR ("SetSecurityDescriptorDacl failed");
      goto finish;
    }

  /* open logfile as stdout/stderr for soon-to-be-spawned subprocess */
  log_handle = CreateFile (log_path[config],
			   GENERIC_WRITE,
			   FILE_SHARE_READ,
			   &sa,
			   append ? OPEN_ALWAYS : CREATE_ALWAYS,
			   FILE_ATTRIBUTE_NORMAL,
			   NULL);

  if (log_handle == INVALID_HANDLE_VALUE)
    {
      ERR ("Cannot open logfile: %s", log_path[config]);
      goto finish;
    }

  /* append to logfile? */
  if (append)
    {
      if (SetFilePointer (log_handle, 0, NULL, FILE_END) == INVALID_SET_FILE_POINTER)
	{
	  ERR ("Cannot seek to end of logfile: %s", log_path[config]);
	  goto finish;
	}
    }

  /* fill in STARTUPINFO struct */
  GetStartupInfo(&start_info);
  start_info.cb = sizeof(start_info);
  start_info.dwFlags = STARTF_USESTDHANDLES|STARTF_USESHOWWINDOW;
  start_info.wShowWindow = SW_HIDE;
  start_info.hStdInput = GetStdHandle(STD_INPUT_HANDLE);
  start_info.hStdOutput = start_info.hStdError = log_handle;

  /* create an OpenVPN process for one config file */
  if (!CreateProcess(exe_path,
		     command_line,
		     NULL,
		     NULL,
		     TRUE,
		     priority | CREATE_NEW_CONSOLE,
		     NULL,
		     config_dir,
		     &start_info,
		     &proc_info))
    {
      ERR ("CreateProcess failed, exe='%s' cmdline='%s' dir='%s'",
          exe_path,
          command_line,
          config_dir);
      goto finish;
    }

  /* close unneeded handles */
  Sleep (1000); /* try to prevent race if we close logfile
		   handle before child process DUPs it */
  if(!CloseHandle (proc_info.hThread) || !CloseHandle (log_handle))
    {
      ERR ("CloseHandle failed");
      goto finish;
    }

  ps_info[config].hProcess=proc_info.hProcess;
  ps_info[config].config=config;


  hThread = CreateThread(NULL, 0, 
            (LPTHREAD_START_ROUTINE) WatchOpenVPNProcess,
            &ps_info[config],  // pass process handle
            0, &IDThread); 
  if (hThread == NULL) 
    {
    // Error exit.
    ERR ("CreateThread to watch openvpn process Failed.");
    goto finish;
  }



finish:
  return; 
}


void StopOpenVPN(int config)
{
  connect_status[config]=false;
  
  if (exit_event[config]) {
    SetEvent(exit_event[config]);
    CloseHandle (exit_event[config]);
  }
  exit_event[config] = NULL;
  SetMenuStatus(config, false);
}

void StopAllOpenVPN()
{
  int i;

  for(i=0; i < num_configs; i++) {
    if(connect_status[i]==true)
      StopOpenVPN(i);
  }
}

void WatchOpenVPNProcess(struct process_info *ps_info)
{
  DWORD exitcode;
  char conn_name[200];
  char msg[250];

  /* Cut of extention from config filename. */
  strncpy(conn_name, config_files[ps_info->config], sizeof(conn_name));
  conn_name[strlen(conn_name) - (strlen(ext_string)+1)]=0;

  /* Wait for a second to see if the process is still alive. */
  if (WaitForSingleObject (ps_info->hProcess, 2000) == WAIT_TIMEOUT) {
    connect_status[ps_info->config]=true;
    SetMenuStatus(ps_info->config, true);

    mysnprintf(msg, "Connection established to %s.", conn_name);
    MessageBox(NULL, msg, "OpenVPN", MB_OK);
   
    /* wait for our openvpn processes to terminate */
    WaitForSingleObject (ps_info->hProcess, -1);

    if(connect_status[ps_info->config] == true) {	/* Process died outside our control */
      ERR("Connection to %s was terminated.", conn_name); 
      CloseHandle(exit_event[ps_info->config]);
      exit_event[ps_info->config] = NULL;    
      connect_status[ps_info->config]=false;
      SetMenuStatus(ps_info->config, false);
    }
  }
  else {
    /* OpenVPN failed to start */
    ERR("Connecting to %s has failed.", conn_name);
    CloseHandle(exit_event[ps_info->config]);
    exit_event[ps_info->config] = NULL;    
  }
}

